Improve doxygen description of mood.c.
[paraslash.git] / sched.c
diff --git a/sched.c b/sched.c
index cad298b839b83a3f28e70912b1c85201ed589bd5..1db9169b38b511918cc711fc28dcb667e1242e5f 100644 (file)
--- a/sched.c
+++ b/sched.c
@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2006-2014 Andre Noll <maan@systemlinux.org>
+ * Copyright (C) 2006 Andre Noll <maan@tuebingen.mpg.de>
  *
  * Licensed under the GPL v2. For licencing details see COPYING.
  */
@@ -7,7 +7,6 @@
 /** \file sched.c Paraslash's scheduling functions. */
 
 #include <regex.h>
-#include <assert.h>
 
 #include "para.h"
 #include "ipc.h"
 #include "time.h"
 #include "error.h"
 
+/**
+ * The possible states of a task.
+ *
+ * In addition to the states listed here, a task may also enter zombie state.
+ * This happens when its ->post_select function returns negative, the ->status
+ * field is then set to this return value. Such tasks are not scheduled any
+ * more (i.e. ->pre_select() and ->post_select() are no longer called), but
+ * they stay on the scheduler task list until \ref task_reap() or
+ * \ref sched_shutdown() is called.
+ */
+enum task_status {
+       /** Task has been reaped and may be removed from the task list. */
+       TS_DEAD,
+       /** Task is active. */
+       TS_RUNNING,
+};
+
 struct task {
-       /** Copied from the task_info struct during task_register(). */
-       void (*pre_select)(struct sched *s, struct task *t);
-       /** Copied from the task_info struct during task_register(). */
-       int (*post_select)(struct sched *s, struct task *t);
-       /** Whether this task is active (>=0) or in error state (<0). */
-       int error;
+       /** A copy of the task name supplied when the task was registered. */
+       char *name;
+       /** Copied during task_register(). */
+       struct task_info info;
+       /* TS_RUNNING, TS_DEAD, or zombie (negative value). */
+       int status;
        /** Position of the task in the task list of the scheduler. */
        struct list_head node;
-       /** The task name supplied when the task was registered(). */
-       char status[255];
        /** If less than zero, the task was notified by another task. */
        int notification;
-       /** True if task is in error state and exit status has been queried. */
-       bool dead;
-       /** Usually a pointer to the struct containing this task. */
-       void *context;
 };
 
 static struct timeval now_struct;
-struct timeval *now = &now_struct;
-
-static inline bool timeout_is_zero(struct sched *s)
-{
-       struct timeval *tv = &s->select_timeout;
-       return tv->tv_sec == 0 && tv->tv_usec == 0;
-}
+const struct timeval *now = &now_struct;
 
 static void sched_preselect(struct sched *s)
 {
        struct task *t, *tmp;
 
        list_for_each_entry_safe(t, tmp, &s->task_list, node) {
-               if (t->error < 0)
+               if (t->status < 0)
                        continue;
                if (t->notification != 0)
                        sched_min_delay(s);
-               if (t->pre_select)
-                       t->pre_select(s, t);
+               if (t->info.pre_select)
+                       t->info.pre_select(s, t->info.context);
        }
 }
 
 static void unlink_and_free_task(struct task *t)
 {
-       PARA_INFO_LOG("freeing task %s\n", t->status);
+       PARA_INFO_LOG("freeing task %s\n", t->name);
        list_del(&t->node);
+       free(t->name);
        free(t);
 }
 
 //#define SCHED_DEBUG 1
 static inline void call_post_select(struct sched *s, struct task *t)
 {
+       int ret;
+
 #ifndef SCHED_DEBUG
-       t->error = t->post_select(s, t);
+       ret = t->info.post_select(s, t->info.context);
 #else
        struct timeval t1, t2, diff;
        unsigned long pst;
 
        clock_get_realtime(&t1);
-       t->error = t->post_select(s, t);
+       ret = t->info.post_select(s, t->info.context);
        clock_get_realtime(&t2);
        tv_diff(&t1, &t2, &diff);
        pst = tv2ms(&diff);
        if (pst > 50)
                PARA_WARNING_LOG("%s: post_select time: %lums\n",
-                       t->status, pst);
+                       t->name, pst);
 #endif
+       t->status = ret < 0? ret : TS_RUNNING;
 }
 
 static unsigned sched_post_select(struct sched *s)
@@ -93,15 +101,14 @@ static unsigned sched_post_select(struct sched *s)
        unsigned num_running_tasks = 0;
 
        list_for_each_entry_safe(t, tmp, &s->task_list, node) {
-               if (t->error < 0) {
-                       if (t->dead) /* task has been reaped */
-                               unlink_and_free_task(t);
-                       continue;
+               if (t->status == TS_DEAD) /* task has been reaped */
+                       unlink_and_free_task(t);
+               else if (t->status == TS_RUNNING) {
+                       call_post_select(s, t); /* sets t->status */
+                       t->notification = 0;
+                       if (t->status == TS_RUNNING)
+                               num_running_tasks++;
                }
-               call_post_select(s, t);
-               t->notification = 0;
-               if (t->error >= 0)
-                       num_running_tasks++;
        }
        return num_running_tasks;
 }
@@ -119,7 +126,7 @@ static unsigned sched_post_select(struct sched *s)
  * \return Zero if no more tasks are left in the task list, negative if the
  * select function returned an error.
  *
- * \sa \ref task, \ref now.
+ * \sa \ref now.
  */
 int schedule(struct sched *s)
 {
@@ -133,7 +140,7 @@ again:
        FD_ZERO(&s->wfds);
        s->select_timeout = s->default_timeout;
        s->max_fileno = -1;
-       clock_get_realtime(now);
+       clock_get_realtime(&now_struct);
        sched_preselect(s);
        ret = s->select_function(s->max_fileno + 1, &s->rfds, &s->wfds,
                &s->select_timeout);
@@ -149,7 +156,7 @@ again:
                FD_ZERO(&s->rfds);
                FD_ZERO(&s->wfds);
        }
-       clock_get_realtime(now);
+       clock_get_realtime(&now_struct);
        num_running_tasks = sched_post_select(s);
        if (num_running_tasks == 0)
                return 0;
@@ -178,27 +185,28 @@ again:
 int task_reap(struct task **tptr)
 {
        struct task *t;
+       int ret;
 
        if (!tptr)
                return 0;
        t = *tptr;
        if (!t)
                return 0;
-       if (t->error >= 0)
-               return 0;
-       if (t->dead) /* will be freed in sched_post_select() */
+       if (t->status >= 0)
                return 0;
+       ret = t->status;
        /*
         * With list_for_each_entry_safe() it is only safe to remove the
         * _current_ list item. Since we are being called from the loop in
         * schedule() via some task's ->post_select() function, freeing the
         * given task here would result in use-after-free bugs in schedule().
* So we only set t->dead which tells schedule() to free the task in
* the next iteration of its loop.
         */
-       t->dead = true;
+       t->status = TS_DEAD;
+
        *tptr = NULL;
-       return t->error;
+       return ret;
 }
 
 /**
@@ -213,10 +221,10 @@ void sched_shutdown(struct sched *s)
        struct task *t, *tmp;
 
        list_for_each_entry_safe(t, tmp, &s->task_list, node) {
-               if (t->error >= 0)
+               if (t->status == TS_RUNNING)
                        /* The task list should contain only terminated tasks. */
                        PARA_WARNING_LOG("shutting down running task %s\n",
-                               t->status);
+                               t->name);
                unlink_and_free_task(t);
        }
 }
@@ -239,31 +247,14 @@ struct task *task_register(struct task_info *info, struct sched *s)
        if (!s->task_list.next)
                INIT_LIST_HEAD(&s->task_list);
 
-       snprintf(t->status, sizeof(t->status) - 1, "%s", info->name);
-       t->status[sizeof(t->status) - 1] = '\0';
+       t->info = *info;
+       t->name = para_strdup(info->name);
        t->notification = 0;
-       t->error = 0;
-       t->dead = false;
-       t->pre_select = info->pre_select;
-       t->post_select = info->post_select;
-       t->context = info->context;
+       t->status = TS_RUNNING;
        list_add_tail(&t->node, &s->task_list);
        return t;
 }
 
-/**
- * Obtain the context pointer of a task.
- *
- * \param t Return this task's context pointer.
- *
- * \return A pointer to the memory location specified previously as \a
- * task_info->context when the task was registered with \ref task_register().
- */
-void *task_context(struct task *t)
-{
-       return t->context;
-}
-
 /**
  * Get the list of all registered tasks.
  *
@@ -282,8 +273,9 @@ char *get_task_list(struct sched *s)
        list_for_each_entry_safe(t, tmp, &s->task_list, node) {
                char *tmp_msg;
                tmp_msg = make_message("%s%p\t%s\t%s\n", msg? msg : "", t,
-                       t->error < 0? (t->dead? "dead" : "zombie") : "running",
-                       t->status);
+                       t->status == TS_DEAD? "dead" :
+                               (t->status == TS_RUNNING? "running" : "zombie"),
+                       t->name);
                free(msg);
                msg = tmp_msg;
        }
@@ -314,7 +306,7 @@ void task_notify(struct task *t, int err)
        assert(err > 0);
        if (t->notification == -err) /* ignore subsequent notifications */
                return;
-       PARA_INFO_LOG("notifying task %s: %s\n", t->status, para_strerror(err));
+       PARA_INFO_LOG("notifying task %s: %s\n", t->name, para_strerror(err));
        t->notification = -err;
 }
 
@@ -346,11 +338,11 @@ int task_status(const struct task *t)
 {
        if (!t)
                return 0;
-       if (t->dead)
+       if (t->status == TS_DEAD) /* pretend dead tasks don't exist */
                return 0;
-       if (t->error >= 0)
+       if (t->status == TS_RUNNING)
                return 1;
-       return t->error;
+       return t->status;
 }
 
 /**
@@ -178,27 +185,28 @@ again:
         * schedule() via some task's ->post_select() function, freeing the
         * given task here would result in use-after-free bugs in schedule().
-        * So we only set t->dead which tells schedule() to free the task in
-        * the next iteration of its loop.
+        * So we only set the task status to TS_DEAD which tells schedule() to
+        * free the task in the next iteration of its loop.
         */